// one line to give the program's name and an idea of what it does.
// Copyright (C) 2005  name of Borja Snchez Zamorano
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

using System;
using System.Drawing;
using System.Windows.Forms;

namespace MonoCalendar.MonthCalendar
{
	public enum MonthCalendarStyle { Day, Week, Month}
	
	public class MonthCalendar : System.Windows.Forms.Control {
		Bitmap _offBmp = null;
		Pen pen1;
		SolidBrush sb1;
		uint firstDayOfWeek;
		DateTime firstDate, firstDateStyle, lastDateStyle;
		DateTime date;
		int[] limits = new int[7];
		int visibleHeight;
		int oldHeight;
		bool sizing = false;
		int monthsViewed = 1;
		int daysByWeek = 7;
		MonthCalendarStyle style;
		MonoCalendar.MonthCalendar.Button buttonLess, buttonMore, buttonToday;
		
		public MonthCalendar()
		{
			pen1 = new Pen(Color.Empty);
			sb1 = new SolidBrush(Color.Empty);
			this.SetStyle(ControlStyles.DoubleBuffer, true);
			this.SetStyle(ControlStyles.UserPaint, true);
			this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
			this.SetStyle(ControlStyles.UserMouse, true);
			this.SetStyle(ControlStyles.ResizeRedraw, false);
			this.SetStyle(ControlStyles.Opaque, true);
			
			System.Globalization.DateTimeFormatInfo dtfi = new System.Globalization.DateTimeFormatInfo();
			firstDayOfWeek = (uint) dtfi.FirstDayOfWeek;
			
			buttonMore = new MonoCalendar.MonthCalendar.Button();
			buttonToday = new MonoCalendar.MonthCalendar.Button();
			buttonLess = new MonoCalendar.MonthCalendar.Button();
			buttonLess.Type = 1;
			buttonMore.Type = 3;
			buttonToday.Type = 2;
			
			this.Controls.Add(buttonMore);
			this.Controls.Add(buttonToday);
			this.Controls.Add(buttonLess);
			buttonToday.Click += new System.EventHandler(ButtonTodayClick);
			buttonMore.Click += new System.EventHandler(ButtonMoreClick);
			buttonLess.Click += new System.EventHandler(ButtonLessClick);
			
			this.Resize += new System.EventHandler(this.MonthCalendarResize);
			this.MouseDown += new System.Windows.Forms.MouseEventHandler(this.MonthCalendarMouseDown);
			this.Paint += new System.Windows.Forms.PaintEventHandler(this.MonthCalendarPaint);
			
			date = DateTime.Now;
		}
		
		void MonthCalendarResize(object sender, System.EventArgs e)
		{
			if ((this.ClientRectangle.Width==0) || (this.ClientRectangle.Height==0)) return;
			
			int ancho = this.ClientRectangle.Width - 14;
			for (int i=0; i<=6; i++) {
				limits[i] = (ancho*(i+1)) / 7;
			}
			int x1, x2, x, width;
			x1 = (this.ClientRectangle.Width/3)+1;
			x2 = this.ClientRectangle.Width*2/3;
			x = 0;
			width = x1;
			buttonLess.SetBounds(x, 0, width, 17);
			x = x1;
			width = x2-x;
			buttonToday.SetBounds(x, 0, width , 17);
			x = x2;
			width = this.ClientRectangle.Width - x;
			buttonMore.SetBounds(x, 0, width, 17);
			
			Repaint2();
		}
		
		void MonthCalendarPaint(object sender, System.Windows.Forms.PaintEventArgs e)
		{
			e.Graphics.DrawImage(_offBmp, 0,0);
		}
		
		int isClickInDate(DateTime date, int y, int mouseX, int mouseY)
		{
			int i;
			int pos = y + 15;
			
			DateTime first = date;
			
			for (i=0; i<=6; i++) {
				if ((uint) first.AddDays(i).DayOfWeek == firstDayOfWeek)
					break;
			}
			
			first = first.AddDays(-(first.Day-1));
			for (i=0; i<=6; i++) {
				if ((uint)first.AddDays(i).DayOfWeek == firstDayOfWeek)
					break;
			}
			i = ((-i+14) % 7);
			
			pos +=19;
			do {
				
				int inits, ends;
				if (i == 0) {
					inits = 5;
				} else {
					inits = limits[i]-11;
				}
				
				if (i == 6) {
					ends = this.ClientRectangle.Width - 5;
				} else {
					ends = limits[i]+8;
				}
				
				if ((pos <= mouseY) && ((pos+18) >= mouseY) && (inits<=mouseX) && (ends>=mouseX)) {
					Date = first;
					OnDateChanged();
					return -1;
				}
				first = first.AddDays(1);
				if (first.Day == 1)
					break;
				i++;
				if (i>6) {
					pos += 18;
					i = 0;
				}
			} while (true);
			
			return (pos+26)-y;
		}
		
		void MonthCalendarMouseDown(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			if (e.Button == System.Windows.Forms.MouseButtons.Left) {
				int x = 23;
				DateTime dateAux = firstDate;
				int i = 0;
				int oldx = 0;
				while (i < monthsViewed) {
					i++;
					oldx = isClickInDate(dateAux, x, e.X, e.Y);
					if (oldx == -1) return;
					x += oldx;
					dateAux = dateAux.AddMonths(1);
				}
			}
		}
		
		void Repaint2()
		{
			if ((this.ClientRectangle.Width<1) || (this.ClientRectangle.Height<1)) return;
			if ((_offBmp == null) || (_offBmp.Height != this.ClientRectangle.Height) || (_offBmp.Width != this.ClientRectangle.Width)) {
				if (_offBmp != null) _offBmp.Dispose();
				_offBmp = new Bitmap( this.ClientRectangle.Width, this.ClientRectangle.Height);
			}
			Graphics _g = Graphics.FromImage( _offBmp );
//			_g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;
			_g.Clear(Color.White);
			PaintWidget(_g);
			_g.Dispose();
			Paint2();
		}
		
		void Paint2()
		{
			if ((_offBmp == null) || (_offBmp.Height != this.ClientRectangle.Height) || (_offBmp.Width != this.ClientRectangle.Width)) {
				Repaint2();
				return;
			}
			if ((this.ClientRectangle.Width<1) || (this.ClientRectangle.Height<1)) return;
			
			IntPtr hwnd = new IntPtr();
			hwnd = this.Handle;
			Graphics drawingSurface = Graphics.FromHwnd(hwnd);
			drawingSurface.DrawImage(_offBmp, 0, 0);
			drawingSurface.Dispose();
		}
		
		public void UpdateCalendar()
		{
			RecalculateDateLimits();
			Repaint2();
		}
		
		void RecalculateDateLimits()
		{
			System.DateTime first, last;
			
			switch(style) {
				case MonthCalendarStyle.Week:
					first = date.AddDays(0);
					for (int i=0; i<7; i++) {
						first = date.AddDays(-i);
						if (daysByWeek==5) {
							if (first.DayOfWeek == DayOfWeek.Monday)
								break;
						} else {
							if ((uint) first.DayOfWeek == firstDayOfWeek)
								break;
						}
					}
					last = first.AddDays(daysByWeek-1);
					break;
				case MonthCalendarStyle.Month:
					first = date.AddDays(-(date.Day - 1));
					last = first.AddMonths(1);
					last = last.AddDays(-last.Day);
					break;
				default:
					first = last = date;
					break;
			}
			
			last = last.AddDays(1).AddSeconds(-1);
			
			firstDateStyle = first;
			lastDateStyle = last;
		}
		
		void PaintBG(Graphics g, int pos, int begin, int end, Color color)
		{
			sb1.Color = color;
			pen1.Color = color;
			
			int inits, ends;
			if (begin == 0) {
				inits = 10;
				g.FillRectangle(sb1, 8, pos, 2, 14);
				g.FillRectangle(sb1, 6, pos+1, 2, 12);
				g.DrawLine(pen1, 5, pos+3, 5, pos + 10);
				g.DrawLine(pen1, 4, pos+5, 4, pos + 8);
			} else {
				inits = limits[begin]-11; // 10
			}
			
			if (end == 6) {
				ends = this.ClientRectangle.Width - 10;
				g.FillRectangle(sb1, this.ClientRectangle.Width-10, pos, 2, 14);
				g.FillRectangle(sb1, this.ClientRectangle.Width-8, pos+1, 2, 12);
				g.DrawLine(pen1, this.ClientRectangle.Width-6, pos+3, this.ClientRectangle.Width-6, pos + 10);
				g.DrawLine(pen1, this.ClientRectangle.Width-5, pos+5, this.ClientRectangle.Width-5, pos + 8);
			} else {
				ends = limits[end]+8;
			}
			
			g.FillRectangle(sb1, inits, pos-1, ends - inits, 16);
		}
		
		int PaintMonth(Graphics g, DateTime date, int y)
		{
			int i;
			int pos = y;
			
			DateTime first = date;
			Font font1 = new Font("Tahoma", 8, FontStyle.Bold);
			StringFormat align1 = new StringFormat();
			align1.Alignment = StringAlignment.Center;
			sb1.Color = Color.FromArgb(255, 0, 0, 0);
			
			g.DrawString(first.ToString("MMMM yyyy"), font1, sb1, this.ClientRectangle.Width/2, pos, align1);
			
			for (i=0; i<=6; i++) {
				if ((uint) first.AddDays(i).DayOfWeek == firstDayOfWeek)
					break;
			}
			
			pos +=15;
			for (int j=0; j<=6; j++) {
				g.DrawString(first.AddDays(i+j).ToString("ddd").ToCharArray()[0].ToString().ToUpper(), font1, sb1, limits[j], pos, align1);
			}
			font1.Dispose();
			font1 = new Font("Tahoma", 8);
			align1.Alignment = StringAlignment.Far;
			
			first = first.AddDays(-(first.Day-1));
			for (i=0; i<=6; i++) {
				if ((uint)first.AddDays(i).DayOfWeek == firstDayOfWeek)
					break;
			}
			i = ((-i+14) % 7);
			
			pos +=19;
			bool paintBG = false;
			do {
				if (!paintBG) {
					PaintBG(g, pos, 0, 6, Color.FromArgb(255, 230, 230, 230));
					
					paintBG = true;
				}
				
				if ((firstDateStyle <= first) && (lastDateStyle >= first)) {
					PaintBG(g, pos, i, i, Color.FromArgb(255, 158, 158, 158));
					sb1.Color = Color.FromArgb(255, 255, 255, 255);
				} else {
					sb1.Color = Color.FromArgb(255, 0, 0, 0);
				}
				
				if (first.ToString("yyyyMMdd") == DateTime.Now.ToString("yyyyMMdd")) {
					Color aux = sb1.Color;
					PaintBG(g, pos, i, i, Color.FromArgb(255, 94, 163, 237));
					sb1.Color = aux;
				}
				
				g.DrawString(first.Day.ToString(), font1, sb1, limits[i]+7, pos, align1);
				
				first = first.AddDays(1);
				if (first.Day == 1)
					break;
				i++;
				if (i>6) {
					paintBG = false;
					pos += 18;
					i = 0;
				}
			} while (true);
			
			font1.Dispose();
			return (pos+26)-y;
		}
		
		void PaintWidget(Graphics g)
		{
			int x = 23;
			int old = 0, all = 150 + 23;;
			DateTime dateAux = firstDate;
			int i = 0;
			int oldx = 0;
			while (x <= this.ClientRectangle.Height) {
				i++;
				old = all;
				all = (i-1)*132 + 150 + 23;
				oldx = x;
				x += PaintMonth(g, dateAux, x);
				dateAux = dateAux.AddMonths(1);
			}
			
			if ((oldx!=0) && (!sizing)) {
				sb1.Color = Color.FromArgb(255, 255, 255, 255);
				g.FillRectangle(sb1, 0, oldx, this.ClientRectangle.Width, this.ClientRectangle.Height - oldx);
			}
			
			oldHeight = visibleHeight;
			
			if (sizing) {
				if (System.Math.Abs(this.ClientRectangle.Height - old) < System.Math.Abs(this.ClientRectangle.Height - all)) {
					visibleHeight = old;
				} else {
					visibleHeight = all;
				}
			} else {
				visibleHeight = old;
			}
			
			MonthsViewed = i - 1;
			if ((oldHeight != visibleHeight) && (!sizing)) {
				OnSizeHeightChanged();
			}
		}

		public int MonthsViewed {
			get {
				return monthsViewed;
			}
			set {
				if (value<=0)
					monthsViewed = 1;
				else
					monthsViewed = value;
			}
		}
		
		void ButtonTodayClick (object sender, System.EventArgs e) {
			Date = DateTime.Now;
			OnDateChanged();
		}
		
		void ButtonMoreClick (object sender, System.EventArgs e) {
			firstDate = firstDate.AddMonths(1);
			Repaint2();
		}
		
		void ButtonLessClick (object sender, System.EventArgs e) {
			firstDate = firstDate.AddMonths(-1);
			Repaint2();
		}
		
		public uint FirstDayOfWeek {
			get {
				return firstDayOfWeek;
			}
			set {
				firstDayOfWeek = value;
				UpdateCalendar();
			}
		}
		
		public int VisibleHeight {
			get {
				return visibleHeight;
			}
			set {
				visibleHeight = value;
			}
		}
		
		public DateTime Date {
			get {
				return date;
			}
			set {
				date = value;
				date = new DateTime(date.Year, date.Month, date.Day);
				
				DateTime lastDate = firstDate.AddMonths(monthsViewed).AddDays(-1);
				if ((date<firstDate) || (date>lastDate)) {
					firstDate = date.AddDays(-(date.Day-1));
				}
				switch(style) {
					case MonthCalendarStyle.Day:
						if (date.ToString("yyyyMMdd") == lastDate.ToString("yyyyMMdd")) {
							firstDate = date.AddDays(-(date.Day-1));
						}
						break;
					case MonthCalendarStyle.Month:
						if (date.ToString("yyyyMM") == lastDate.ToString("yyyyMM")) {
							firstDate = date.AddDays(-(date.Day-1));
						}
						break;
					case MonthCalendarStyle.Week:
						{
							DateTime aux = date.AddDays(1);
							while ((uint)aux.DayOfWeek != firstDayOfWeek) {
								aux = aux.AddDays(1);
							}
							aux = aux.AddDays(-1);
							if (aux >= lastDate) {
								firstDate = date.AddDays(-(date.Day-1));
							}
						}
						break;
				}
				firstDate = firstDate.AddDays(-(firstDate.Day-1));
				firstDate = new DateTime(firstDate.Year, firstDate.Month, firstDate.Day);
				
				UpdateCalendar();
				if (firstDate > firstDateStyle) {
					firstDate = firstDateStyle.AddDays(-(firstDate.Day-1));
					firstDate = new DateTime(firstDate.Year, firstDate.Month, firstDate.Day);
					UpdateCalendar();
					Date = Date;
				}
			}
		}
		
		public int OldHeight {
			get {
				return oldHeight;
			}
			set {
				oldHeight = value;
			}
		}
		
		public bool Sizing {
			get {
				return sizing;
			}
			set {
				sizing = value;
				Repaint2();
			}
		}
		
		public MonthCalendarStyle Style {
			get {
				return style;
			}
			set {
				style = value;
				UpdateCalendar();
			}
		}
		
		public int DaysByWeek {
			get {
				return daysByWeek;
			}
			set {
				daysByWeek = value;
				UpdateCalendar();
			}
		}

		protected virtual void OnDateChanged()
		{
			if (DateChanged != null) DateChanged(this, System.EventArgs.Empty);
		}
		
		protected virtual void OnSizeHeightChanged()
		{
			if (SizeHeightChanged != null) SizeHeightChanged(this, System.EventArgs.Empty);
		}
		
		public event EventHandler DateChanged;
		
		public event EventHandler SizeHeightChanged;
	}
}
